import re

from .base import BasePhoneticsAlgorithm


class Metaphone(BasePhoneticsAlgorithm):
    def __init__(self, compress_ending=False):
        """
        Initialization of Metaphone object
        :param compress_ending: not used
        """
        self.compress_ending = compress_ending

    _deaf_regex = re.compile(r'', re.I)
    _vowels_table = str.maketrans('', '')

    def _deaf_consonants_letters(self, word):
        return word

    def _compress_ending(self, word):
        return word

    def _apply_metaphone_algorithm(self, word):
        word = self._reduce_seq(word)
        word = word.translate(self._vowels_table)
        word = self._deaf_consonants_letters(word)
        if self.compress_ending:
            word = self._compress_ending(word)
        return word.upper()

    def transform(self, word):
        return self._apply_metaphone_algorithm(word)


class RussianMetaphone(Metaphone):
    _vowels = 'аэиоуыеёюя'
    _deaf_consonants = str.maketrans('бздвг', 'пстфк')
    _vowels_table = str.maketrans('аяоыиеёэюу', 'ААААИИИИУУ')

    _j_vowel_regex = re.compile(r'[ий][ео]', re.I)

    _replacement_vowel_map = {
        re.compile(r'(^|ъ|ь|' + r'|'.join(_vowels) + r')(я)', re.I): 'йа',
        re.compile(r'(^|ъ|ь|' + r'|'.join(_vowels) + r')(ю)', re.I): 'йу',
        re.compile(r'(^|ъ|ь|' + r'|'.join(_vowels) + r')(е)', re.I): 'йе',
        re.compile(r'(^|ъ|ь|' + r'|'.join(_vowels) + r')(ё)', re.I): 'йо',
        re.compile(r'[ъь]', re.I): ''
    }

    _replacement_map = {
        re.compile(r'сч', re.I): r'щ',
        re.compile(r'([тсзжцчшщ])([жцчшщ])', re.I): r'\2',
        re.compile(r'(с)(т)([лнц])', re.I): r'\1\3',
        re.compile(r'(н)([тд])(ств)', re.I): r'\1\3',
        re.compile(r'([нс])([тд])(ск)', re.I): r'\1\3',
        re.compile(r'(р)([гк])(ск)', re.I): r'\1\3',
        re.compile(r'(р)(д)([чц])', re.I): r'\1\3',
        re.compile(r'(з)(д)([нц])', re.I): r'\1\3',
        re.compile(r'(ль|н)(д)(ш)', re.I): r'\1\3',
        re.compile(r'(н)(т)(г)', re.I): r'\1\3',
        re.compile(r'(в)(ств)', re.I): r'\2',
        re.compile(r'(л)(нц)', re.I): r'\2',
        re.compile(r'([дт][сц])', re.I): 'ц'
    }

    def __init__(self, compress_ending=False, reduce_phonemes=False):
        super().__init__(compress_ending)
        self.reduce_phonemes = reduce_phonemes

    def _replace_j_vowels(self, word):
        for replace, result in self._replacement_vowel_map.items():
            word = replace.sub(result, word)
        return self._j_vowel_regex.sub('и', word)

    def _reduce_phonemes(self, word):
        for replace, result in self._replacement_map.items():
            word = replace.sub(result, word)
        return word

    def _compress_ending(self, word):
        return word

    def _deaf_consonants_letters(self, word):
        res = []
        for i, letter in enumerate(word):
            if i == len(word) - 1 or \
                    letter in self._deaf_consonants and (word[i + 1] not in 'лмнр' or word[i + 1] not in self._vowels):
                res += [letter.translate(self._deaf_consonants)]
            else:
                res += [letter]
        return ''.join(res)

    def transform(self, word):
        if self.reduce_phonemes:
            word = self._reduce_phonemes(word)
        word = self._replace_j_vowels(word)
        return self._apply_metaphone_algorithm(word)


class FinnishMetaphone(Metaphone):
    _vowels = 'aäeioöuy'
    _deaf_consonants = str.maketrans('bvdg', 'pftk')
    _vowels_table = str.maketrans('aäeioöuy', 'ÄÄIIIIУУ')

    _z_replacement = re.compile(r'z', re.I)
    _q_replacement = re.compile(r'q', re.I)
    _w_replacement = re.compile(r'w', re.I)
    _x_replacement = re.compile(r'x', re.I)

    def _deaf_consonants_letters(self, word):
        res = []
        for i, letter in enumerate(word):
            if i == len(word) - 1 or \
                    letter in self._deaf_consonants and (word[i + 1] not in 'lmnr' or word[i + 1] not in self._vowels):
                res += [letter.translate(self._deaf_consonants)]
            else:
                res += [letter]
        return ''.join(res)

    def transform(self, word):
        word = self._z_replacement.sub('ts', word)
        word = self._q_replacement.sub('k', word)
        word = self._w_replacement.sub('v', word)
        word = self._x_replacement.sub('ks', word)
        return self._apply_metaphone_algorithm(word)



import editdistance
from .base import BasePhoneticsAlgorithm
from .soundex import Soundex


class PhoneticsException(Exception):
    def __init__(self, msg):
        self.msg = msg


class PhoneticsDistance:
    @staticmethod
    def _hamming(word1, word2):
        if len(word1) != len(word2):
            raise PhoneticsException('For Hamming distance words should be the same length!')
        return sum(a != b for a, b in zip(word1, word2))

    @staticmethod
    def _levenstein(word1, word2):
        return editdistance.eval(word1, word2)

    _distance_metric = {
        'levenstein': _levenstein.__func__,
        'hamming': _hamming.__func__,
    }

    def __init__(self, phonetics, metric_name='levenstein', metrics=None):
        """
        Init a distance object
        :param phonetics: an object of BasePhoneticsAlgorithm class
        :param metric_name: distance function name, optional, default is Levenstein distance
        :param metrics: another distance function, optional
        """
        assert isinstance(phonetics, BasePhoneticsAlgorithm)
        self.phonetics = phonetics

        if not metrics:
            self.metrics = self._distance_metric[metric_name]
        else:
            self.metrics = metrics

    def distance(self, word1, word2):
        """
        Compute the distance between phonetics codes
        :param word1: first original word
        :param word2: second original word
        :return: distance value
        """
        w1, w2 = self.phonetics.transform(word1), self.phonetics.transform(word2)
        if isinstance(self.phonetics, Soundex):
            return self.metrics(w1, w2) if self.phonetics.is_delete_first_letter() else self.metrics(w1[1:], w2[1:])
        return self.metrics(w1, w2)
